_TapeMeasure = {}
local TM = _TapeMeasure

local testMode = false

TM.context = UI.CreateContext("TapeMeasure")

local tx_lastupdate = 0
local tx_firstupdate = 0
local ui_setup = false
local _TMVisible = true
local dispmode = "ALL"

local _iud = Inspect.Unit.Detail

local _allunits = {
	"group01", "group02", "group03", "group04", "group05",
	"group06", "group07", "group08", "group09", "group10",
	"group11", "group12", "group13", "group14", "group15",
	"group16", "group17", "group18", "group19", "group20"
	}

TM.frames = {
	header = {},
	options = {},
	target = {},
	focus = {},
	members= {}
}

local _unitDetails = {}
local _unitDist = {}

local function UnitDetails(u)
	_unitDetails = {}
	if type(u) == "table" then
		for k,v in pairs(u) do
			_unitDetails[k] = {
				name=string.format("Player%02d", k),
				coordX = _unitDist[k].x,
				coordY = _unitDist[k].y,
				coordZ = _unitDist[k].z,
				id = k
			}
		end
--		TM.SlashHandler("test")
	elseif u == "player" then
		_unitDetails = {
			name=u,
			coordX = 0,
			coordY = 0,
			coordZ = 0
		}
	elseif u == "focus" then
		_unitDetails = {
			name=u,
			coordX = 9,
			coordY = 18,
			coordZ = 21
		}
	elseif u == "player.target" then
		local d = Inspect.Unit.Detail(u)
		if d then
			_unitDetails = {
				name=u,
				coordX = 12,
				coordY = 7,
				coordZ = 3
			}
		else
			_unitDetails = nil
		end
	end
	return _unitDetails
end


local function MergeTable(o,n)
	for k,v in pairs(n) do
		if type(v) == "table" then
			if not o[k] then
				o[k] = {}				
			end
			MergeTable(o[k], n[k])
		else
			if not o[k] then
				o[k] = v
			end
		end
	end
end

local default_settings = {
	icon_x = UIParent:GetWidth()/2,
	icon_y = UIParent:GetHeight()/2,
	targ_x = (UIParent:GetWidth()/2)+50,
	targ_y = UIParent:GetHeight()/2,
	focu_x = (UIParent:GetWidth()/2)+100,
	focu_y = UIParent:GetHeight()/2,
	x = 0,
	y = 0,
	docked = "Y",
	distance = 6,
	alwaysshow = "Y",
	sortorder = "N",
	showfocus = "Y"
	}

local frVis = 0
local sqrt = math.sqrt
	
function TM.UpdateUI()
	local pdtl = _iud("player")
	local dtls
	local tdtl = _iud("player.target")
	local fdtl = _iud("focus")
	local dx
	local frUsed = 0
	local dist = tonumber(TapeMeasure_Settings.distance) or 0
	
	if pdtl.coordX then
		if tdtl then
			if tdtl.coordX then
				dx = sqrt(((pdtl.coordX - tdtl.coordX)^2)+((pdtl.coordY - tdtl.coordY)^2)+((pdtl.coordZ - tdtl.coordZ)^2))
				TM.frames.target.text:SetText(string.format("Distance to %s: %.2fm", tdtl.name, dx))
				TM.frames.target.icon:SetVisible(true)
			else
				if TapeMeasure_Settings.alwaysshow=="Y" then
					TM.frames.target.text:SetText("Distance to target: NO TARGET")
					TM.frames.target.icon:SetVisible(true)
				else
					TM.frames.target.icon:SetVisible(false)
				end
			end
		else
			if TapeMeasure_Settings.alwaysshow=="Y" then
				TM.frames.target.text:SetText("Distance to target: NO TARGET")
				TM.frames.target.icon:SetVisible(true)
			else
				TM.frames.target.icon:SetVisible(false)
			end		
		end
		if fdtl then
			if fdtl.coordX then
				dx = sqrt(((pdtl.coordX - fdtl.coordX)^2)+((pdtl.coordY - fdtl.coordY)^2)+((pdtl.coordZ - fdtl.coordZ)^2))
				TM.frames.focus.text:SetText(string.format("Distance to %s: %.2fm", fdtl.name, dx))
				TM.frames.focus.icon:SetVisible(true and (TapeMeasure_Settings.showfocus=="Y"))
			else
				if TapeMeasure_Settings.alwaysshow=="Y" then
					TM.frames.focus.text:SetText("Distance to focus: NO TARGET")
					TM.frames.focus.icon:SetVisible(true and (TapeMeasure_Settings.showfocus=="Y"))
				else
					TM.frames.focus.icon:SetVisible(false)
				end
			end
		else
			if TapeMeasure_Settings.alwaysshow=="Y" then
				TM.frames.focus.text:SetText("Distance to focus: NO TARGET")
				TM.frames.focus.icon:SetVisible(true and (TapeMeasure_Settings.showfocus=="Y"))
			else
				TM.frames.focus.icon:SetVisible(false)
			end
		end
		dtls = _iud(_allunits)
		local dxtab = {}
		for k,v in pairs(dtls) do
			if v.coordX and pdtl.coordX then
				dx = sqrt(((pdtl.coordX - v.coordX)^2)+((pdtl.coordY - v.coordY)^2)+((pdtl.coordZ - v.coordZ)^2))
				if dx<dist and (v.id ~= pdtl.id) then
					table.insert(dxtab, {n=v.name, d=dx})
				end
			end
		end
		if TapeMeasure_Settings.sortorder == "N" then
			table.sort(dxtab, function(a,b) return a.d<b.d end)
		else
			table.sort(dxtab, function(a,b) return a.d>b.d end)
		end
		for _,vx in ipairs(dxtab) do			
			frUsed = frUsed+1
			TM.frames.members[frUsed].name:SetText(vx.n)
			TM.frames.members[frUsed].dist:SetText(string.format("%.2fm",vx.d))
		end
	else
		-- player dead!
	end
--	print("frUsed = "..frUsed..", frVis = "..frVis)
	if frUsed > frVis then
		for x=1, frUsed do
			TM.frames.members[x].name:SetVisible(true)
		end
		frVis = frUsed
	elseif frUsed < frVis then
		for x=frUsed+1, frVis do
			TM.frames.members[x].name:SetVisible(false)
		end
		frVis = frUsed
	end
end	

function TM.SetupUI()
	TapeMeasure_Settings = TapeMeasure_Settings or {}
	MergeTable(TapeMeasure_Settings, default_settings)
	
	TM.frames.header.icon = UI.CreateFrame("Texture", "icon", TM.context)
	TM.frames.header.icon:SetHeight(36)
	TM.frames.header.icon:SetWidth(36)
	TM.frames.header.icon:SetTexture("TapeMeasure", "TapeMeasure.png")
	TM.frames.header.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", TapeMeasure_Settings.x, TapeMeasure_Settings.y)
	
	TM.frames.target.icon = UI.CreateFrame("Texture", "icon", TM.context)
	TM.frames.target.icon:SetHeight(24)
	TM.frames.target.icon:SetWidth(24)
	TM.frames.target.icon:SetTexture("TapeMeasure", "target.png")
	
	TM.frames.target.text = UI.CreateFrame("Text", "dttarget", TM.frames.target.icon)
	TM.frames.target.text:SetFontSize(14)
	TM.frames.target.text:SetText("Distance to target: ")
	TM.frames.target.text:SetPoint("CENTERLEFT", TM.frames.target.icon, "CENTERRIGHT", 0,0)

	TM.frames.focus.icon = UI.CreateFrame("Texture", "icon", TM.context)
	TM.frames.focus.icon:SetHeight(24)
	TM.frames.focus.icon:SetWidth(24)
	TM.frames.focus.icon:SetTexture("TapeMeasure", "focus.png")

	TM.frames.focus.text = UI.CreateFrame("Text", "dtfocus", TM.frames.focus.icon)
	TM.frames.focus.text:SetFontSize(14)
	TM.frames.focus.text:SetText("Distance to focus: ")
	TM.frames.focus.text:SetPoint("CENTERLEFT", TM.frames.focus.icon, "CENTERRIGHT", 0,0)

	if TapeMeasure_Settings.docked == "Y" then
		TM.frames.target.icon:SetPoint("BOTTOMLEFT", TM.frames.header.icon, "TOPRIGHT", 0,0)
		TM.frames.focus.icon:SetPoint("BOTTOMLEFT", TM.frames.target.icon, "TOPLEFT", 0,0)
	else
		TM.frames.target.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", TapeMeasure_Settings.targ_x, TapeMeasure_Settings.targ_y)
		TM.frames.focus.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", TapeMeasure_Settings.focu_x, TapeMeasure_Settings.focu_y)
	end
	
	function TM.frames.target.icon.Event:LeftDown()
		if TapeMeasure_Settings.docked == "N" then
			self.mouseLDown = true
			local md = Inspect.Mouse()
			self.sx = md.x - TM.frames.target.icon:GetLeft()
			self.sy = md.y - TM.frames.target.icon:GetTop()
		else
			self.mouseLDown = false
		end
	end
	
	function TM.frames.target.icon.Event:MouseMove()
		if self.mouseLDown then
			local nx, ny
			local md = Inspect.Mouse()
			nx = md.x - self.sx
			ny = md.y - self.sy
			TM.frames.target.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx, ny)
			TapeMeasure_Settings.targ_x = nx
			TapeMeasure_Settings.targ_y = ny
		end
	end

	function TM.frames.target.icon.Event:LeftUp()
		self.mouseLDown = false
	end
	
	function TM.frames.focus.icon.Event:LeftDown()
		if TapeMeasure_Settings.docked == "N" then
			self.mouseLDown = true
			local md = Inspect.Mouse()
			self.sx = md.x - TM.frames.focus.icon:GetLeft()
			self.sy = md.y - TM.frames.focus.icon:GetTop()
		else
			self.mouseLDown = false
		end
	end
	
	function TM.frames.focus.icon.Event:MouseMove()
		if self.mouseLDown then
			local nx, ny
			local md = Inspect.Mouse()
			nx = md.x - self.sx
			ny = md.y - self.sy
			TM.frames.focus.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx, ny)
			TapeMeasure_Settings.focu_x = nx
			TapeMeasure_Settings.focu_y = ny
		end
	end

	function TM.frames.focus.icon.Event:LeftUp()
		self.mouseLDown = false
	end
	
	TM.frames.header.title = UI.CreateFrame("Text", "header", TM.context)
	TM.frames.header.title:SetText("Players closer than: ")
	TM.frames.header.title:SetFontSize(18)
	TM.frames.header.title:SetLayer(1)
	TM.frames.header.title:SetPoint("TOPLEFT", TM.frames.header.icon, "TOPRIGHT", 0,0)

	coroutine.yield()

	local tx_click = 0
	
	function TM.frames.header.icon.Event:LeftDown()
		self.mouseLDown = true
		local md = Inspect.Mouse()
		self.sx = md.x - TM.frames.header.icon:GetLeft()
		self.sy = md.y - TM.frames.header.icon:GetTop()
	end
	
	function TM.frames.header.icon.Event:MouseMove()
		if self.mouseLDown then
			local nx, ny
			local md = Inspect.Mouse()
			nx = md.x - self.sx
			ny = md.y - self.sy
			TM.frames.header.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", nx, ny)
			TapeMeasure_Settings.x = nx
			TapeMeasure_Settings.y = ny
		end
	end

	function TM.frames.header.icon.Event:LeftUp()
		self.mouseLDown = false
	end
	
	function TM.frames.header.icon.Event:RightClick()
		if TapeMeasure_Settings.docked == "Y" then
			TapeMeasure_Settings.docked = "N"
			print("Undocking target/focus distance frames")
			TM.frames.target.icon:ClearPoint("BOTTOMLEFT")
			TM.frames.target.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", TapeMeasure_Settings.targ_x, TapeMeasure_Settings.targ_y)
			TM.frames.focus.icon:ClearPoint("BOTTOMLEFT")
			TM.frames.focus.icon:SetPoint("TOPLEFT", UIParent, "TOPLEFT", TapeMeasure_Settings.focu_x, TapeMeasure_Settings.focu_y)
		else
			TapeMeasure_Settings.docked = "Y"
			print("Docking target/focus distance frames")
			TM.frames.target.icon:ClearPoint("TOPLEFT")
			TM.frames.target.icon:SetPoint("BOTTOMLEFT", TM.frames.header.icon, "TOPRIGHT", 0,0)
			TM.frames.focus.icon:ClearPoint("TOPLEFT")
			TM.frames.focus.icon:SetPoint("BOTTOMLEFT", TM.frames.target.icon, "TOPLEFT", 0,0)
		end
	end

	TM.frames.options.dist = UI.CreateFrame("RiftTextfield", "dist", TM.frames.header.title)
	TM.frames.options.dist:SetHeight(TM.frames.header.title:GetHeight())
	TM.frames.options.dist:SetWidth(TM.frames.header.icon:GetWidth()*1.5)
	TM.frames.options.dist:SetBackgroundColor(0,0,0,0.25)
	TM.frames.options.dist:SetText(string.format("%.2f", (TapeMeasure_Settings.distance or 0)))
	TM.frames.options.dist:SetPoint("TOPLEFT", TM.frames.header.title, "TOPRIGHT", 0,0)
	
	function TM.frames.options.dist.Event.TextfieldChange()
		local txt = TM.frames.options.dist:GetText()
		if txt == "" then txt = 0 end
		TapeMeasure_Settings.distance = txt
	end

	coroutine.yield()
	
	for x=1,20 do
		TM.frames.members[x] = {}
		if x==1 then
			TM.frames.members[x].name = UI.CreateFrame("Text", "member", TM.frames.header.title)
			TM.frames.members[x].name:SetPoint("TOPLEFT", TM.frames.header.title, "BOTTOMLEFT", 0,0)
		else
			TM.frames.members[x].name = UI.CreateFrame("Text", "member", TM.frames.members[x-1].name)
			TM.frames.members[x].name:SetPoint("TOPLEFT", TM.frames.members[x-1].name, "BOTTOMLEFT", 0,0)
		end
		TM.frames.members[x].name:SetFontSize(12)		
		TM.frames.members[x].name:SetText(string.format("PLAYER: %02d", x))
		TM.frames.members[x].dist = UI.CreateFrame("Text", "member", TM.frames.members[x].name)
		TM.frames.members[x].dist:SetFontSize(12)
		TM.frames.members[x].dist:SetPoint("TOPLEFT", TM.frames.members[x].name, "TOPRIGHT", 0,0)
		TM.frames.members[x].dist:SetText(string.format("DIST%02d", x))
		TM.frames.members[x].name:SetVisible(false)
		if x%5 == 0 then
			coroutine.yield()
		end
	end
	ui_setup = true
end	

function TM.AddonLoaded(a)
	if a == "TapeMeasure" then
		local dx = Inspect.Addon.Detail(a)
		print("v"..dx.toc.Version.." loaded")
	end
end

function TM.SlashHandler(args)
	local cmd, arg = string.match(args, "(%a+) ?(.*)")
	if cmd == "test" then
		testMode = not testMode
		print("Test mode: "..tostring(testMode))
		if testMode then
			_iud = UnitDetails
		else
			_iud = Inspect.Unit.Detail
		end
	elseif cmd == "run" then
		_test()
	elseif cmd == "dist" then
		local dist = string.match(arg, "(%d+.?%d*)") or 0
		TapeMeasure_Settings.distance = dist
		TM.frames.options.dist:SetText(string.format("%.2f", dist))
		print("dist = "..dist)
	elseif cmd == "alwaysshow" then
		if TapeMeasure_Settings.alwaysshow=="Y" then
			TapeMeasure_Settings.alwaysshow="N"
			print("Showing frames when no target: NO")
		else
			TapeMeasure_Settings.alwaysshow="Y"
			print("Showing frames when no target: YES")
		end
	elseif cmd == "focus" then
		if TapeMeasure_Settings.showfocus == "Y" then
			TapeMeasure_Settings.showfocus = "N"
			print("Showing focus frame: NO")
		else
			TapeMeasure_Settings.showfocus = "Y"
			print("Showing focus frame: YES")
		end
	elseif cmd == "sort" then
		if TapeMeasure_Settings.sortorder == "N" then
			TapeMeasure_Settings.sortorder = "F"
			print("Sort order: Far to Near")
		else
			TapeMeasure_Settings.sortorder = "N"
			print("Sort order: Near to Far")
		end
	else
		print("OPTIONS:")
		print("dist N.NN - sets warning distance to N.NN")
		print("alwaysshow - toggles display of frames when no target selected")
		print("focus - toggles display of focus distance")
		print("sort - toggles sort order of raid list")
		print("test - toggles test mode")
	end
end

local updatecnt = 0

function TM.OnUpdate()
	local tx_curupdate = Inspect.Time.Frame()
	if tx_firstupdate == 0 then
		tx_firstupdate = tx_curupdate
		TM.coUISetup = coroutine.create(TM.SetupUI)
	elseif (tx_curupdate - tx_lastupdate) > 0.25 then
		if ui_setup then
			TM.UpdateUI()
		end
		updatecnt = updatecnt + 1
		tx_lastupdate = tx_curupdate
	end
	if (updatecnt >= 8) and (not ui_setup) then
		local s = coroutine.status(TM.coUISetup)
		if s == "suspended" then
			coroutine.resume(TM.coUISetup)
		end
	end
end

table.insert(Event.Addon.Load.End, {TM.AddonLoaded, "TapeMeasure", "Loaded" })
table.insert(Command.Slash.Register("tape"), { TM.SlashHandler, "TapeMeasure", "SlashHandler" })
table.insert(Event.System.Update.Begin, { TM.OnUpdate, "TapeMeasure", "OnUpdate" })

for x=1,20 do
	local k1 = math.random(5*x)
	local k2 = math.random(x)
	local k3 = math.random(5*x)
	table.insert(_unitDist, {x=k1, y=k2, z=k3})
end